/*
* Arduino source code for my K-9 robot.
* Copyright (C) 2017 Colin Leroy-Mira <colin@colino.net>
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <avr/wdt.h>
#include <Servo.h>
#include <math.h>

#define DEBUG_RC 0
#define DEBUG_MOTOR 0
#define DEBUG_EARS 0
#define DEBUG_NOSE 1
#define DEBUG_NECK 0
#define DEBUG_TAIL 0
#define DEBUG_VOLTAGE 0

int debug_counter = 0;

/* inputs */
#define RC_CHANNEL1_PIN 9
#define RC_CHANNEL2_PIN 8
#define RC_CHANNEL3_PIN 7
#define RC_CHANNEL4_PIN 6

bool rc_calibrated = false;
long rc_signal_lost = 0;
int calibration_samples = 0;
long ch1_neutral = 0;
long ch2_neutral = 0;
int ch1_max_amplitude = 300;
int ch2_max_amplitude = 300;
int ch1;
int ch2;
int ch3;
int ch4;

/* command/control values */
#define MIN_RC_THRESHOLD 30.0f

#define IDLE_DELAY 60
#define SHUTDOWN_DELAY 10

float throttle_percent = 0;
float steering_percent = 0;
int nose_position = 0;

float neck_middle = 98.0f;
float neck_position = neck_middle;
float neck_min = 60.0f;
float neck_max = 135.0f;
int neck_way = 0;

int tail_middle = 90;
int tail_position = tail_middle;
int tail_min = 60;
int tail_max = 120;
int tail_way = 0;

float left_motor_percent = 0;
float right_motor_percent = 0;

int nose_servos_stop = 87;

/* outputs */
#define POWER_PIN 38

#define MOTOR_LEFT_DIR 5
#define MOTOR_LEFT_PWM 4
#define MOTOR_RIGHT_DIR 3
#define MOTOR_RIGHT_PWM 2

#define MOTOR_EAR_LEFT 22
#define MOTOR_EAR_RIGHT 24

#define MOTOR_NOSE_LEFT 26
#define MOTOR_NOSE_RIGHT 28

#define NOSE_REAR_STOP 30
#define NOSE_FRONT_STOP 32

#define NECK_SERVO 36

#define TAIL_SERVO 43

#define LEDS_1 45
#define LEDS_2 47
#define LEDS_3 49
#define LEDS_4 51

/* Voltage reader using voltage diviser (pin, resistors values) */
#define VOLTAGE_PIN A0
#define R1 10000.0
#define R2 9100.0
#define CUTOFF_VOLTAGE 6.00f

Servo noseMotorLeft;
Servo noseMotorRight;
Servo earMotorLeft;
Servo earMotorRight;
Servo neckMotor;
Servo tailMotor;

void do_reset() {
  /* Install the watchdog */
  wdt_enable(WDTO_15MS);

  /* Trigger the watchdog */
  while(true) {
   
  } 
}

long now() {
  return millis() / 1000;  
}

void reset_all() {
  digitalWrite(MOTOR_LEFT_PWM, HIGH); 
  digitalWrite(MOTOR_RIGHT_PWM, HIGH); 

  analogWrite(MOTOR_LEFT_DIR, 0);  
  analogWrite(MOTOR_RIGHT_DIR, 0);  

  noseMotorLeft.write(nose_servos_stop);
  noseMotorRight.write(nose_servos_stop);

  earMotorLeft.write(90);
  earMotorRight.write(90);
  neckMotor.write((int)neck_middle);
  tailMotor.write(tail_middle);

  digitalWrite(LEDS_1, LOW);  
  digitalWrite(LEDS_2, LOW);  
  digitalWrite(LEDS_3, LOW);  
  digitalWrite(LEDS_4, LOW);  

  steering_percent = 0;
  throttle_percent = 0;
  nose_position = 0;
  neck_position = neck_middle;
  neck_way = 0;
}

int off = 0;

void cut_power() {
  if (off) {
    return;
  }
  Serial.println("Shutting down");
  
  digitalWrite(POWER_PIN, LOW);  
  off = 1;
}

void setup() {
  pinMode(POWER_PIN, OUTPUT);  
  digitalWrite(POWER_PIN, HIGH);  

  Serial.begin(115200);
  
  pinMode(RC_CHANNEL1_PIN, INPUT);
  pinMode(RC_CHANNEL2_PIN, INPUT);
  pinMode(RC_CHANNEL3_PIN, INPUT);
  pinMode(RC_CHANNEL4_PIN, INPUT);

  pinMode(VOLTAGE_PIN, INPUT);

  pinMode(NOSE_REAR_STOP, INPUT_PULLUP);
  pinMode(NOSE_FRONT_STOP, INPUT_PULLUP);

  pinMode(MOTOR_LEFT_DIR, OUTPUT);
  pinMode(MOTOR_LEFT_PWM, OUTPUT);
  pinMode(MOTOR_RIGHT_DIR, OUTPUT);
  pinMode(MOTOR_RIGHT_PWM, OUTPUT);

  pinMode(LEDS_1, OUTPUT);  
  pinMode(LEDS_2, OUTPUT);  
  pinMode(LEDS_3, OUTPUT);  
  pinMode(LEDS_4, OUTPUT);  

  noseMotorLeft.attach(MOTOR_NOSE_LEFT);
  noseMotorRight.attach(MOTOR_NOSE_RIGHT);

  earMotorLeft.attach(MOTOR_EAR_LEFT);
  earMotorRight.attach(MOTOR_EAR_RIGHT);
  neckMotor.attach(NECK_SERVO);
  tailMotor.attach(TAIL_SERVO);

  reset_all();
  
  randomSeed(analogRead(12));

  Serial.println("Waiting for RC signal calibration");
}

void read_rc() {
  int i, j;

  ch1 = pulseIn(RC_CHANNEL1_PIN, HIGH, 25000); // Read the pulse width of 
  ch2 = pulseIn(RC_CHANNEL2_PIN, HIGH, 25000); // each channel  
  ch3 = pulseIn(RC_CHANNEL3_PIN, HIGH, 25000);
  ch4 = pulseIn(RC_CHANNEL4_PIN, HIGH, 25000);
}

void try_calibrating_rc() {
  if (ch1 > 400) {
    /* TX is on, we can store values for calibration */
    ch1_neutral += ch1;
    ch2_neutral += ch2;
    calibration_samples++;
  }
  if (calibration_samples >= 20) {
    ch1_neutral = ch1_neutral / calibration_samples;
    ch2_neutral = ch2_neutral / calibration_samples;
    
    Serial.println("Calibrated.");
    rc_calibrated = true;
  }
}

int rcounter = 0;

void standardize_rc() {
  rcounter++;
  if (rcounter % 8 == 0) {
    rcounter = 0;
  } else {
    return;
  }
  
  if (ch1 < 100) {
    /* Safe guard for RC loss */
    if (rc_calibrated == true) {
      Serial.println("RC signal lost.");
      rc_calibrated = false;
      calibration_samples = 0;
      ch1_neutral = 0;
      ch2_neutral = 0;
      rc_signal_lost = now();
      /* Stop everything */
      reset_all();
    }
    return;
  }
  ch1 -= ch1_neutral;
  ch2 -= ch2_neutral;
  
  if (ch1 < -ch1_max_amplitude) {
    ch1 = -ch1_max_amplitude;
  }
  if (ch1 > ch1_max_amplitude) {
    ch1 = ch1_max_amplitude;
  }
  if (ch2 < -ch2_max_amplitude) {
    ch2 = -ch2_max_amplitude;
  }
  if (ch2 > ch2_max_amplitude) {
    ch2 = ch2_max_amplitude;
  }

  /* steering */
  float s_percent = ((float)ch1  * (float)100) / (float)ch1_max_amplitude;
  /* throttle */
  float t_percent = ((float)ch2  * (float)100) / (float)ch2_max_amplitude;
  /* nose */
  if (ch3 > 1500) {
    nose_position = 0;
  } else {
    nose_position = 1;
  }

  /* neck */
  if (ch4 > 1500) {
    neck_way = -1;
  } else if (ch4 < 1100) {
    neck_way = 1;
  } else {
    neck_way = 0;
  }

  /* Shift from (MIN_RC_THRESHOLD to 100) to (0 to 100) - steering */
  if (abs(s_percent) < MIN_RC_THRESHOLD) {
    s_percent = 0;
  } else if (s_percent > 0) {
    s_percent -= MIN_RC_THRESHOLD;
  } else {
    s_percent += MIN_RC_THRESHOLD;
  }
  s_percent = (s_percent * 100) / (100 - MIN_RC_THRESHOLD);

  /* Same for throttle */
  if (abs(t_percent) < MIN_RC_THRESHOLD) {
    t_percent = 0;
  } else if (t_percent > 0) {
    t_percent -= MIN_RC_THRESHOLD;
  } else {
    t_percent += MIN_RC_THRESHOLD;
  }
  t_percent = (t_percent * 100) / (100 - MIN_RC_THRESHOLD);

  /* Set final values - steering is inverted due to RC controller */
  steering_percent = -s_percent;
  throttle_percent = t_percent;
}

long last_move = 0;

void check_for_powerdown() {
  /* Auto-shutdown in case we forget to turn it off */
  if (now() - last_move > IDLE_DELAY) {
    Serial.print("Idling since ");
    Serial.println(now() - last_move);
    cut_power();
  }

  /* Auto-shutdown if RC signal lost */
  if (rc_calibrated == false && now() - rc_signal_lost > SHUTDOWN_DELAY) {
    Serial.print("Lost RC signal since ");
    Serial.println(now() - last_move);
    cut_power();
  }

  /* Battery power check 
   * https://en.wikipedia.org/wiki/Voltage_divider
   * vout is what the pin reads,
   * vin is what the battery provides 
   */
  float vout = analogRead(VOLTAGE_PIN);
  vout = (vout * 5.0f) / 1024.0f;
  
  float vin = vout / (R2 / (R1+R2));

#if DEBUG_VOLTAGE
  Serial.print("Voltage: vout = ");
  Serial.print(vout);
  Serial.print(", vin = ");
  Serial.println(vin);
#endif
  if (vin < CUTOFF_VOLTAGE && !off) {
    Serial.print("Low voltage : ");
    Serial.print(vin);
    Serial.println("V");
    cut_power();
  }
}

#if DEBUG_MOTOR
float steering_factor = 0;
#endif

void check_idle() {
  /* We just booted, initialize last_move */
  if(last_move == 0) {
    last_move = now();
    return;
  }
  /* We're moving, update last move */
  if(abs(throttle_percent) > MIN_RC_THRESHOLD || abs(steering_percent) > MIN_RC_THRESHOLD) {
    last_move = now();
    return;
  }
  check_for_powerdown();
}

/* Derive motors power and direction from steering/throttle
 * input steering_percent + throttle_percent variables
 * output left/right_motor_percent variables
 * */
void adapt_rc() {
    float magnitude = min(100, sqrt(pow(throttle_percent,2) + pow(steering_percent,2)));
    float myAngle = atan2(throttle_percent,steering_percent);
    myAngle = myAngle - (3.1415f/4);
    
    left_motor_percent = magnitude * cos(myAngle);
    right_motor_percent = magnitude * sin(myAngle); 

#if DEBUG_MOTOR
  Serial.print("left_motor_percent ");
  Serial.print(left_motor_percent);
  Serial.print(" right_motor_percent ");
  Serial.print(right_motor_percent);
  Serial.println("");
#endif
}

/* Apply current settings to the tracks motors
 * input left_motor_percent, right_motor_percent
 */
void run_motors() {
  int left_way;
  int right_way;
  int left_speed;
  int right_speed;

  /* Motor controller takes an 8-bit value for speed */
  left_speed  = abs((255 * left_motor_percent) / 100);
  right_speed = abs((255 * right_motor_percent) / 100);

  /* don't bother if it's close to zero */
  if (left_speed < 25) {
    left_speed = 0;
  }
  if (right_speed < 25) {
    right_speed = 0;
  }

  /* Set motors direction */
  if (left_speed ==  0 && right_speed == 0) {
    left_way = HIGH;
    right_way = HIGH;
  } else {
    left_way = (left_motor_percent < 0) ? HIGH : LOW;
    right_way = (right_motor_percent < 0) ? HIGH : LOW;
  }

#if DEBUG_MOTOR
  Serial.print("d1 ");
  Serial.print(left_way);
  Serial.print(" d2 ");
  Serial.print(right_way);
  Serial.print(" s1 ");
  Serial.print(left_speed);
  Serial.print(" s2 ");
  Serial.print(right_speed);
  Serial.println("");
#endif

  /* Update motor controller directions */
  digitalWrite(MOTOR_LEFT_PWM, left_way); 
  digitalWrite(MOTOR_RIGHT_PWM, right_way); 

  /* Update motor controller speed */
  analogWrite(MOTOR_LEFT_DIR, left_speed);  
  analogWrite(MOTOR_RIGHT_DIR, right_speed);  
}

void debug() {
  if (debug_counter % 10 == 0) {
#if DEBUG_RC
    if (abs(steering_percent) > 5.0f || abs(throttle_percent) > 5.0f) {
      Serial.print("Read - Steering / Throttle : ");
      Serial.print(steering_percent);
      Serial.print(" / ");
      Serial.print(throttle_percent);
      Serial.print("\n");
    }
#endif
#if DEBUG_MOTOR
    if (abs(left_motor_percent) > 5.0f || abs(right_motor_percent) > 5.0f) {
      Serial.print("Motors left / right / steering redux: ");
      Serial.print(left_motor_percent);
      Serial.print(" / ");
      Serial.print(right_motor_percent);
      Serial.print(" / ");
      Serial.print(steering_factor);
      Serial.print("\n");
    }
#endif
    debug_counter = 0;
  }
  debug_counter++;
}

long last_ear_program = 0;
int move_delay = 3;
int num_ear_move = 0;
int ear_delay = 0;
int ear_same_way = 0;

int left_ear_pos = 0;
int right_ear_pos = 0;
#define MIN_EAR_POS -40
#define MAX_EAR_POS +40

long last_ear_move = 0;

void move_ears() {
  /* If ears aren't moving but they did not very long
   * ago, do nothing */
  if (num_ear_move == 0 && now() - last_ear_program < move_delay) {
     return;
  }

  last_ear_program = now();
  
  /* skip sometimes (1 out of 4 times) */
  if ((num_ear_move == 0 && random(1, 100) % 4 == 0) || rc_calibrated == false) {
#if DEBUG_EARS
    if (rc_calibrated == true) {
      Serial.println("skipping ear move");
    }
#endif
    return;
  }

  /* Randomize ear movement */
  if (num_ear_move == 0) {
    num_ear_move = random(1, 6);
    move_delay = random(6, 10);
    ear_delay = random(1, 6) * 200;
    ear_same_way = random(0, 2);
#if DEBUG_EARS
    Serial.print("num ear moves to do = ");
    Serial.print(num_ear_move);
    Serial.print(",delay till next ear sequence = ");
    Serial.println(move_delay);
#endif
  }

  if (millis() - last_ear_move < ear_delay) {
    return;
  }
  
  last_ear_move = millis();
  ear_delay = random(1, 6) * 200;

  left_ear_pos  = random((MIN_EAR_POS / 10), (MAX_EAR_POS / 10) + 1);
  left_ear_pos *= 10;

  if (ear_same_way != 0) {
    right_ear_pos = left_ear_pos;
  } else {
    right_ear_pos = -left_ear_pos;
  }
  
  left_ear_pos += 90;
  right_ear_pos += 90;

#if DEBUG_EARS
  Serial.print("setting ears for move ");
  Serial.print(num_ear_move);
  Serial.print(" : ");
  Serial.print(left_ear_pos);
  Serial.print(" / ");
  Serial.println(right_ear_pos);
#endif

  earMotorLeft.write(left_ear_pos);
  earMotorRight.write(right_ear_pos);
  
  num_ear_move--;
}

int nose_way = 0;
void move_nose() {
  /* dont' move nose before RC is ready */
  if (!rc_calibrated) {
    return;
  }

  int val;

  /* Do we want to retract or extend nose ? */
  if (nose_position == 0 && digitalRead(NOSE_REAR_STOP)) {
    /* must keep retracting */
#if DEBUG_NOSE
    Serial.println("nose not in yet");
#endif
    nose_way = -1;
  } else if (nose_position == 1 && digitalRead(NOSE_FRONT_STOP)) {
    /* must keep extending */
#if DEBUG_NOSE
    Serial.println("nose not out yet");
#endif
    nose_way = 1;
  } else {
    /* Nothing to move */
    nose_way = 0;
  }

  /* Send order to servos */
  if (nose_way == 0) {
    noseMotorLeft.write(nose_servos_stop);
    noseMotorRight.write(nose_servos_stop);
  } else if (nose_way == 1) {
    noseMotorLeft.write(180);
    noseMotorRight.write(0);
  } else if (nose_way == -1) {
    noseMotorLeft.write(0);
    noseMotorRight.write(180);
  }
}

void move_neck() {
  if (neck_way == 0) {
    /* do nothing */
  } else if (neck_position + ((float)neck_way / 1.0f) >= neck_min 
          && neck_position + ((float)neck_way / 1.0f) <= neck_max ) {
    neck_position += (float)neck_way / 1.0f;
  }

#if DEBUG_NECK
  Serial.print("moving neck to ");
  Serial.print(neck_position);
  Serial.print (" (");
  Serial.print((int)neck_position);
  Serial.println(")");
#endif

  neckMotor.write((int)neck_position);
}

long last_tail_move = 0;

void move_tail() {
  int new_position;
  if (millis() - last_tail_move < 250) {
    return;
  }

  /* if the tail is not moving */
  if (tail_way == 0) {
    /* 1% chance to move it */
    if (random(1, 100) == 20) {
      /* choose which way randomly */
      tail_way = (random(1, 3) == 1) ? 1 : -1;
#if DEBUG_TAIL
      Serial.println("moving tail");
#endif
    }
  } 
  if (tail_way == 1) {
    new_position = tail_max;
    tail_way = -1;
  } else if (tail_way == -1) {
    new_position = tail_min;
    tail_way = 1;
  } else {
    new_position = tail_middle;
  }

  /* If tail isn't centered, put it back in the middle (1/3rd random chance) */
  if (new_position != tail_middle && tail_way != 0 && random(1, 4) == 1) {
    tail_way = 0;
#if DEBUG_TAIL
    Serial.println("stopping tail");
#endif
  }

  if (new_position != tail_position) {
    tail_position = new_position;
    last_tail_move = millis();
    tailMotor.write((int)tail_position);
  }
}

/* Randomize which LEDs are on every second */
long last_led_change = 0;
void light_leds() {
  if (millis() - last_led_change > 1000) {
    last_led_change = millis();

    digitalWrite(LEDS_1, (random(1,3) == 1) ? HIGH : LOW); 
    digitalWrite(LEDS_2, (random(1,3) == 1) ? HIGH : LOW); 
    digitalWrite(LEDS_3, (random(1,3) == 1) ? HIGH : LOW); 
    digitalWrite(LEDS_4, (random(1,3) == 1) ? HIGH : LOW); 

  }
}

/* Main loop */
void loop() {  
  wdt_reset();
  read_rc();

  /* check whether we're calibrated */
  if(rc_calibrated == false) {
    
    try_calibrating_rc();
    delay(100);
  } else {
   /* game on ! */
    standardize_rc();
    adapt_rc();
    run_motors();
    move_ears();
    move_nose();
    move_neck();
    move_tail();
    light_leds();
  }
  check_idle();
  check_for_powerdown();

  debug();
}
